(****************************************************************************)

(*                               UNIT TCPIP                                 *)

(****************************************************************************)

(*  Zum Dekodieren von TCPIP-Frames im Monitor von MCUT.                    *)

(*                                                                          *)

(* Diese Unit ist Teil des MCUT-Sourcecodes, alle Rechte liegen bei         *)

(* DG4IAD und DG6VJ                                                         *)

(*                                                                          *)

(* dg6vj, dez 95                                                            *)

(****************************************************************************)

UNIT TCPIP;





INTERFACE



(* Dekodieren eines IP-Frames. bergabe NUR des TCP/IP-Teils, NICHTS vom

 * AX25-Protokoll.

 * Siehe RFC-791

 *

 * bergabe/Rckgabe:

 * In Frame steht das Frame mit der PID $CC. Lnge bis zu 255 Bytes.

 * Zurckgegeben wird in Header der komplette extrahierte TCPIP-Header

 * und in BackFrame die Nutzdaten des Frames. Sollte das AX25-Frame

 * grer als 255 Bytes gewesen sein, kann dieses briggebliebene Byte

 * einfach an BackFrame angehngt werden, und schon ist es wieder

 * vollstndig.

 *)

PROCEDURE DekodeFrame(Frame: STRING; VAR Header,BackFrame: STRING);







IMPLEMENTATION



TYPE str20 = STRING[20];

     str4 = STRING[4];



{Hier wird die "Protokoll-Zahl" in einen String umgewandelt. Weitere

 Protokolle knnen einfach in diese CASE-Schleife eingetragen werden}

FUNCTION ProtocolLong(Nr: BYTE): str20;

VAR S: STRING;

BEGIN

  CASE Nr OF

     1: ProtocolLong := 'ICMP';

     2: ProtocolLong := 'IGMP';

     3: ProtocolLong := 'GGP';

     5: ProtocolLong := 'ST';

     6: ProtocolLong := 'TCP';

     7: ProtocolLong := 'UCL';

     8: ProtocolLong := 'EGP';

     9: ProtocolLong := 'IGP';

    17: ProtocolLong := 'UDP';

    21: ProtocolLong := 'PRM';

    27: ProtocolLong := 'RDP';

    28: ProtocolLong := 'IRTP';

    71: ProtocolLong := 'IPCV';

    80: ProtocolLong := 'ISO-IP';

    84: ProtocolLong := 'TTP';

    87: ProtocolLong := 'TCF';

    88: ProtocolLong := 'IGRP';

    ELSE BEGIN

      STR(Nr,S);

      ProtocolLong := S;

    END;

  END;

END;





{Ports "ausschreiben"}

FUNCTION PortLong(Nr: WORD): STRING;

VAR S: STRING;

BEGIN

  CASE Nr OF

         7: PortLong := 'ECHO';

         9: PortLong := 'DISCARD';

        13: PortLong := 'NETTIME';

        20: PortLong := 'FTP-DATA';

        21: PortLong := 'FTP';

        23: PortLong := 'TELNET';

        25: PortLong := 'SMTP';

        37: PortLong := 'TIME';

        53: PortLong := 'DNS';

        70: PortLong := 'GOPHER';

        79: PortLong := 'FINGER';

        80: PortLong := 'WWW';

        87: PortLong := 'TTYLINK';

       109: PortLong := 'POP2';

       110: PortLong := 'POP3';

       119: PortLong := 'NNTP';

      3600: PortLong := 'CONVERS';

      6809: PortLong := 'KaConvers';

      5800..5999: PortLong := 'X10';

      6000..6099: PortLong := 'X11';

    ELSE BEGIN

           STR(Nr,S);

           PortLong := S;

         END;

  END;

END;





{Macht aus der 32-Bit-IP-Adresse (4 Bytes) einen lesbaren String

 (44.130.22.68)}

FUNCTION MakeIPAdr(Adr: str4): STRING;

VAR S: STRING[20];

    Hlp: STRING[3];

BEGIN

  STR(Ord(Adr[1]),S);

  STR(Ord(Adr[2]),Hlp);

  S := S + '.' + hlp;

  STR(Ord(Adr[3]),Hlp);

  S := S + '.' + hlp;

  STR(Ord(Adr[4]),Hlp);

  MakeIPAdr := S + '.' + hlp;

END;





{Dekodiert das TCP-Frame (Transmission Control Protocol) und gibt das

 ausgewertete Frame zurck. Auerdem wird das TCP-Teil aus dem Frame

 herausgeschnitten und der Rest zurckgegeben.



 Siehe RFC-793          }

FUNCTION DecodeTCP_Frame(VAR Frame: STRING): STRING;

VAR Window,CRC,SourcePort,DestiPort: WORD;

    SeqNr,AckNr: LONGINT;

    Offset: BYTE;

    Control: BYTE;

    Hlp,S: STRING;

BEGIN



  {*************************************************************************}

  {****************************** TCP-Header *******************************}

  {*************************************************************************}



{

  TCP Header Format





    0                   1                   2                   3

    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1

   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

   |          Source Port          |       Destination Port        |

   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

   |                        Sequence Number                        |

   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

   |                    Acknowledgment Number                      |

   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

   |  Data |           |U|A|P|R|S|F|                               |

   | Offset| Reserved  |R|C|S|S|Y|I|            Window             |

   |       |           |G|K|H|T|N|N|                               |

   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

   |           Checksum            |         Urgent Pointer        |

   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

   |                    Options                    |    Padding    |

   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

   |                             data                              |

   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

}





  {********************* Source- und Destination-Port **********************}

  SourcePort := Byte(Frame[1])*256 + Byte(Frame[2]);

  DestiPort := Byte(Frame[3])*256 + Byte(Frame[4]);



  {***************************** Sequence-Nr *******************************}

  SeqNr := Byte(Frame[8]) + Byte(Frame[7])*256 + Byte(Frame[6])*65536{ + Byte(Frame[5])*16777216};



  {*************************** Acknowledge-Nr ******************************}

  AckNr := Byte(Frame[12]) + Byte(Frame[11])*256 + Byte(Frame[10])*65536{ + Byte(Frame[9])*16777216};



  {************************* TCP-Frame-Lnge *******************************}

  {** Ab dieser Stelle (* 32 Bit= * 4 Bytes) fangen die Daten an ***********}

  Offset := (Byte(Frame[13]) SHR 4);



  {************************** Control-Bits *********************************}

  {Control Bits:  6 bits (from left to right):



    URG:  Urgent Pointer field significant

    ACK:  Acknowledgment field significant

    PSH:  Push Function

    RST:  Reset the connection

    SYN:  Synchronize sequence numbers

    FIN:  No more data from sender}

  Control := Byte(Frame[14]) AND 63;





  {******************** maximale Window-Gre *****************************}

  Window := Byte(Frame[15])*256 + Byte(Frame[16]);





  {String zusammenstckeln}

  S := 'TCP: '+PortLong(SourcePort) + '->' + PortLong(DestiPort) + ' Seq ';

  STR(SeqNr,Hlp);

  S := S + Hlp;

  IF Control AND 1>0 THEN S := S + ' FIN';

  IF Control AND 2>0 THEN S := S + ' SYN';

  IF Control AND 4>0 THEN S := S + ' RST';

  IF Control AND 8>0 THEN S := S + ' PSH';

  IF Control AND 16>0 THEN S := S + ' ACK';

  IF Control AND 32>0 THEN S := S + ' URG';

  DecodeTCP_Frame := S;

  Frame := COPY(Frame,Offset*4,255);

END;





(* Auswertung von ICMP-Messages (Internet Control Message Protocol)

 * Siehe RFC-792

 *)

FUNCTION DecodeICMP_Frame(VAR Frame: STRING): STRING;

CONST Unreachable : ARRAY[0..5] OF STRING[32] =

                    ('net unreachable','host unreachable','protocol unreachable',

                     'port unreachable','fragmentation needed and DF set','source route failed');

      Exceeded : ARRAY[0..1] OF STRING[40] =

                 ('time-to-live exceeded in transmit','fragment reassembly time exceed');

      Redirect : ARRAY[0..3] OF STRING[45] =

                 ('Redirect for the Network: ',

                  'Redirect for the Host: ',

                  'Redirect for the Type of Service and Network: ',

                  'Redirect for the Type of Service and Host: ');



VAR Typ: BYTE;

    S,Hlp: STRING;

BEGIN

  {*************************************************************************}

  {**************************** ICMP ***************************************}

  {*************************************************************************}



{

   ICMP Router Advertisement Message



       0                   1                   2                   3

       0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1

      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

      |     Type      |     Code      |           Checksum            |

      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

      |   Num Addrs   |Addr Entry Size|           Lifetime            |

      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

      |                       Router Address[1]                       |

      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

      |                      Preference Level[1]                      |

      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

      |                       Router Address[2]                       |

      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

      |                      Preference Level[2]                      |

      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

      |                               .                               |

      |                               .                               |

      |                               .                               |

}

  S := 'ICMP: type ';



  {Typ der Message}

  Typ := Byte(Frame[1]);



  CASE Typ OF

     0: BEGIN {Echo Reply Message}

          S := S + 'Echo Reply id ';

          STR(Byte(Frame[5])*256 + Byte(Frame[6]),Hlp);

          S := S + Hlp + ' seq ';

          STR(Byte(Frame[7])*256 + Byte(Frame[8]),Hlp);

          S := S + Hlp;

          Frame := COPY(Frame,9,255);

        END;

     3: BEGIN {Destination Unreachable Message}

          S := S + Unreachable[Byte(Frame[2])];

          Frame := COPY(Frame,9,255);

        END;

     5: BEGIN {Redirect Message}

          S := S + Redirect[Byte(Frame[2])]+MakeIPAdr(COPY(Frame,5,4));

          Frame := COPY(Frame,9,255);

        END;

     8: BEGIN {Echo Message}

          S := S + 'Echo Request id ';

          STR(Byte(Frame[5])*256 + Byte(Frame[6]),Hlp);

          S := S + Hlp + ' seq ';

          STR(Byte(Frame[7])*256 + Byte(Frame[8]),Hlp);

          S := S + Hlp;

          Frame := COPY(Frame,9,255);

        END;

    11: BEGIN {Time Exceeded Message}

          S := S + Exceeded[Byte(Frame[2])];

          Frame := COPY(Frame,9,255);

        END;

    13: BEGIN {Timestamp Message}

          STR(Byte(Frame[12]) + Byte(Frame[11])*256 + Byte(Frame[10])*65536{ + Byte(Frame[9])*16777216},Hlp);

          S := S + 'Timestamp Message'+#13+'      Originate: '+Hlp;

          Frame := COPY(Frame,21,255);

        END;

    14: BEGIN {Timestamp Reply Message}

          STR(Byte(Frame[12]) + Byte(Frame[11])*256 + Byte(Frame[10])*65536{ + Byte(Frame[9])*16777216},Hlp);

          S := S + 'Timestamp Reply Message'+#13+'      Originate: '+Hlp;

          STR(Byte(Frame[16]) + Byte(Frame[15])*256 + Byte(Frame[14])*65536{ + Byte(Frame[13])*16777216},Hlp);

          S := S + #13 + '      Recieve: '+Hlp;

          STR(Byte(Frame[20]) + Byte(Frame[19])*256 + Byte(Frame[18])*65536{ + Byte(Frame[17])*16777216},Hlp);

          S := S + #13 + '      Transmit: '+Hlp;

          Frame := COPY(Frame,21,255);

        END;

    15: BEGIN {Information Request}

          S := S + 'Information Request ';

          STR(Byte(Frame[5])*256 + Byte(Frame[6]),Hlp);

          S := S + Hlp + ' seq ';

          STR(Byte(Frame[7])*256 + Byte(Frame[8]),Hlp);

          S := S + Hlp;

          Frame := COPY(Frame,9,255);

        END;

    16: BEGIN {Information Request Reply}

          S := S + 'Information Request Reply ';

          STR(Byte(Frame[5])*256 + Byte(Frame[6]),Hlp);

          S := S + Hlp + ' seq ';

          STR(Byte(Frame[7])*256 + Byte(Frame[8]),Hlp);

          S := S + Hlp;

          Frame := COPY(Frame,9,255);

       END;

    ELSE S := S + 'unknown';

  END;



  DecodeICMP_Frame := S;

END;







PROCEDURE DekodeFrame(Frame: STRING; VAR Header,BackFrame: STRING);

VAR SourceAdr,DestiAdr: STRING[4];

    Len: WORD;

    Protocol, IHL, TTL, Flags: BYTE;

    S,Hlp: STRING;

BEGIN

  {*************************************************************************}

  (****************************** IP-Teil **********************************)

  {*************************************************************************}



{

                    Example Internet Datagram Header



    0                   1                   2                   3

    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1

   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

   |Version|  IHL  |Type of Service|          Total Length         |

   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

   |         Identification        |Flags|      Fragment Offset    |

   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

   |  Time to Live |    Protocol   |         Header Checksum       |

   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

   |                       Source Address                          |

   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

   |                    Destination Address                        |

   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

   |                    Options                    |    Padding    |

   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+



                    Example Internet Datagram Header

}







  {*************************** Source-Adresse ******************************}

  SourceAdr := COPY(Frame,13,4);



  {************************ Destination-Adresse ****************************}

  DestiAdr := COPY(Frame,17,4);



  {************************** Lnge des Frames *****************************}

  Len := Byte(Frame[3])*256 + Byte(Frame[4]);



  {****************************** Protokol *********************************}



  { Aus RFC 1060



      Dezimal    Keyword     Protocol                         References

      -------    -------     --------                         ----------

           0                 Reserved                              [JBP]

           1     ICMP        Internet Control Message           [97,JBP]

           2     IGMP        Internet Group Management          [43,JBP]

           3     GGP         Gateway-to-Gateway                  [60,MB]

           4                 Unassigned                            [JBP]

           5     ST          Stream                             [49,JWF]

           6     TCP         Transmission Control              [106,JBP]

           7     UCL         UCL                                    [PK]

           8     EGP         Exterior Gateway Protocol        [123,DLM1]

           9     IGP         any private interior gateway          [JBP]

          10     BBN-RCC-MON BBN RCC Monitoring                    [SGC]

          11     NVP-II      Network Voice Protocol             [22,SC3]

          12     PUP         PUP                               [8,XEROX]

          13     ARGUS       ARGUS                                [RWS4]

          14     EMCON       EMCON                                 [BN7]

          15     XNET        Cross Net Debugger                [56,JFH2]

          16     CHAOS       Chaos                                 [NC3]

          17     UDP         User Datagram                     [104,JBP]

          18     MUX         Multiplexing                       [23,JBP]

          19     DCN-MEAS    DCN Measurement Subsystems           [DLM1]

          20     HMP         Host Monitoring                    [59,RH6]

          21     PRM         Packet Radio Measurement              [ZSU]

          22     XNS-IDP     XEROX NS IDP                    [133,XEROX]

          23     TRUNK-1     Trunk-1                              [BWB6]

          24     TRUNK-2     Trunk-2                              [BWB6]

          25     LEAF-1      Leaf-1                               [BWB6]

          26     LEAF-2      Leaf-2                               [BWB6]

          27     RDP         Reliable Data Protocol            [138,RH6]

          28     IRTP        Internet Reliable Transaction      [79,TXM]

          29     ISO-TP4     ISO Transport Protocol Class 4    [63,RC77]

          30     NETBLT      Bulk Data Transfer Protocol       [20,DDC1]

          31     MFE-NSP     MFE Network Services Protocol    [124,BCH2]

          32     MERIT-INP   MERIT Internodal Protocol             [HWB]

          33     SEP         Sequential Exchange Protocol        [JC120]

          34     3PC         Third Party Connect Protocol         [SAF3]

       35-60                 Unassigned                            [JBP]

          61                 any host internal protocol            [JBP]

          62     CFTP        CFTP                              [50,HCF2]

          63                 any local network                     [JBP]

          64     SAT-EXPAK   SATNET and Backroom EXPAK             [SHB]

          65                 Unassigned                            [JBP]

          66     RVD         MIT Remote Virtual Disk Protocol      [MBG]

          67     IPPC        Internet Pluribus Packet Core         [SHB]

          68                 any distributed file system           [JBP]

          69     SAT-MON     SATNET Monitoring                     [SHB]

          70     VISA        VISA Protocol                        [GXT1]

          71     IPCV        Internet Packet Core Utility          [SHB]

       72-75                 Unassigned                            [JBP]

          76     BR-SAT-MON  Backroom SATNET Monitoring            [SHB]

          77     SUN-ND      SUN ND PROTOCOL-Temporary             [WM3]

          78     WB-MON      WIDEBAND Monitoring                   [SHB]

          79     WB-EXPAK    WIDEBAND EXPAK                        [SHB]

          80     ISO-IP      ISO Internet Protocol                 [MTR]

          81     VMTP        VMTP                                 [DRC3]

          82     SECURE-VMTP SECURE-VMTP                          [DRC3]

          83     VINES       VINES                                 [BXH]

          84     TTP         TTP                                   [JXS]

          85     NSFNET-IGP  NSFNET-IGP                            [HWB]

          86     DGP         Dissimilar Gateway Protocol      [74,ML109]

          87     TCF         TCF                                  [GAL5]

          88     IGRP        IGRP                               [18,GXS]

          89     OSPFIGP     OSPFIGP                           [83,JTM4]

          90     Sprite-RPC  Sprite RPC Protocol               [143,BXW]

          91     LARP        Locus Address Resolution Protocol     [BXH]

       92-254                Unassigned                            [JBP]

          255                Reserved                              [JBP]

  }

  Protocol := Byte(Frame[10]);





  {******************* IHL - Internet Header Length ************************}

  IHL := Byte(Frame[1]);

  IHL := (IHL AND 15) * 4;



  {*************************** TTL - Time To Life **************************}

  TTL := Byte(Frame[9]);





  S := 'IP: ' + MakeIPAdr(SourceAdr)+'->'+MakeIPAdr(DestiAdr)+' ihl:';

  STR(IHL,Hlp);

  S := S + Hlp + ' ttl:';

  STR(TTL,Hlp);

  S := S + Hlp + ' prot:' + ProtocolLong(Protocol);





  {**************** Check, ob das Frame fragmentiert ist ******************}

  Flags := Byte(Frame[8]) SHR 5;

  IF Flags AND 2>0 THEN S := S + ' fragmented';







  {************************** IP-Header abschneiden ************************}

  Frame := COPY(Frame,IHL+1,255);





  {Sooo... Nu gucken, was fr ein Protokoll weiter verwendet wird.}

  Hlp := '';

  CASE Protocol OF

    1: Hlp := DecodeICMP_Frame(Frame); {ICMP - Internet Control Message Protocol}

    6: Hlp := DecodeTCP_Frame(Frame);{TCP - Transmission Control Protocol}

  END;



  IF LENGTH(Hlp)>0 THEN S := S + #13 + Hlp;

  {In S steht der TCPIP-Heade drin -> zurckgeben}

  Header := S;



  {Jetzt sind nur noch Nutzdaten in Frame -> auch zurckgeben}

  BackFrame := Frame;



END;







END.

